# -*- coding: utf-8 -*-

"""
Yuna allows different tables to be serialized using different
serialization formats.  For example, if your needs are simple you might serialize using
JSON; for more space efficiency you might serialize using MesgPack.

Yuna allows different tables to be compressed.  You might skip this,
or use zlib or LZ4 compression, or maybe even zstandard compression.
In service of the above, this file has tables used to look up serialization
or compression.

The reason we look up live, rather than importing everything, is to allow
users to only "pip install" the optional features they need.  If you
never need to write or open a Yuna file with zstandard compression, you
you can skip "pip install zstandard".

When you attempt to open a file, the metadata lists out all the serialization
and compression formats used in the file, using tags (simple strings; for
example, the tag for LZ4 compression is the string "lz4").

If you have an old version of Yuna and you try to open a new Yuna DB file
you may have a lookup failure because a serialization or compression format
is not present in this file.  In that case, try checking for a newer
release of Yuna DB.
"""

from json import dumps as json_dumps
from json import loads as json_loads

from typing import Any, Callable, Dict, Iterator, List, Optional


import lmdb


from .lmdb_util import YUNA_DEFAULT_MAX_DB_SIZE, YUNA_DEFAULT_MAX_TABLES
from .lmdb_util import YUNA_DB_META_KEY, YUNA_FILE_EXTENSION

from .lmdb_util import _lmdb_open, _yuna_get_meta, _yuna_put_meta
from .lmdb_util import _lmdb_table_delete, _lmdb_table_get, _lmdb_table_put


# Create a unique object used to detect if optional arg not provided.
# Can't use None because user might want to provide None.
_YUNA_NOT_PROVIDED = object()


class SerializePlugins:
    def __init__(self,
        serialize: Callable,
        deserialize: Callable,
        init: Optional[Callable] = None,
        options: Optional[Callable] = None,
    ):
        self.serialize = serialize
        self.deserialize = deserialize
        self.init = init
        self.options = options

class CompressPlugins:
    def __init__(self,
        compress: Callable,
        decompress: Callable,
        init: Optional[Callable] = None,
        options: Optional[Callable] = None,
        train: Optional[Callable] = None,
    ):
        self.compress = compress
        self.decompress = decompress
        self.init = init
        self.options = options
        self.train = train


YUNA_SERIALIZE_CACHE: Dict[str, SerializePlugins] = {}


def _not_implemented(*args, **kwargs) -> None:
    raise NotImplemented("not implemented yet")

def _empty_string_key_check(x: str) -> None:
    # LMDB doesn't allow zero-length keys.  For now, raise when one is seen.  TODO: consider adding a workaround.
    if not isinstance(x, int) and x is not None:
        if not x:
            raise ValueError("key cannot be empty string")


INTEGER_KEY = "integer_key"


SERIALIZE_JSON = "json"

def serialize_json(x: Any) -> bytes:
    return json_dumps(x).encode("utf-8")

deserialize_json = json_loads

def _import_json() -> None:
    import json

    plugins = SerializePlugins(
        serialize=serialize_json,
        deserialize=deserialize_json,
    )
    YUNA_SERIALIZE_CACHE[SERIALIZE_JSON] = plugins


SERIALIZE_MSGPACK = "msgpack"


def _import_msgpack() -> None:
    import msgpack

    plugins = SerializePlugins(
        serialize=msgpack.dumps,
        deserialize=msgpack.loads,
    )
    YUNA_SERIALIZE_CACHE[SERIALIZE_MSGPACK] = plugins


SERIALIZE_STR = "str"

def serialize_str(s: str) -> bytes:
    return s.encode('utf-8')

def deserialize_str(bytes_s: bytes) -> str:
    return str(bytes_s, 'utf-8')

def _import_str() -> None:
    # nothing to import; strings are built-in
    plugins = SerializePlugins(
        serialize=serialize_str,
        deserialize=deserialize_str,
    )
    YUNA_SERIALIZE_CACHE[SERIALIZE_STR] = plugins


_SERIALIZE_IMPORT_FUNCTIONS = {
    SERIALIZE_JSON: _import_json,
    SERIALIZE_MSGPACK: _import_msgpack,
    SERIALIZE_STR: _import_str,
}



YUNA_COMPRESS_CACHE: Dict[str, CompressPlugins] = {}


COMPRESS_LZ4 = "lz4"

def _import_lz4() -> None:
    import lz4
    import lz4.block

    plugins = CompressPlugins(
        compress=lz4.block.compress,
        decompress=lz4.block.decompress,
    )
    YUNA_COMPRESS_CACHE[COMPRESS_LZ4] = plugins


COMPRESS_ZLIB = "zlib"

def _import_zlib() -> None:
    import zlib

    plugins = CompressPlugins(
        init=None,
        options=None,  # TODO: add an options function
        compress=zlib.compress,
        decompress=zlib.decompress,
    )
    YUNA_COMPRESS_CACHE[COMPRESS_ZLIB] = plugins


COMPRESS_ZSTD = "zstd"

def _init_zstd():
    raise RuntimeError("init not implemented yet but is coming")
def _options_zstd(*args, **kwargs):
    raise RuntimeError("options not implemented yet but is coming")
def _train_zstd_factory() -> Callable:
    # build the function inside this factory so that it will close over the module reference
    fn_train_dictionary = zstandard.train_dictionary
    def _train_zstd(size: int, samples: List[Any]) -> bytes:
        """
        @size: how many bytes the dictionary should be
        @samples: list of training data records

        Builds a compression dictionary of size @size from data in @samples
        """
        # TODO: see if an iterator works for @samples and update docs if it does
        compression_dictionary = fn_train_dictionary(*args, **kwargs)
        bytes_data = compression_dictionary.as_bytes()
        return bytes_data
    return _train_zstd

def _import_zstd() -> None:
    import zstandard

    plugins = CompressPlugins(
        init=None, # TODO: add init()
        options=None,  # TODO: add an options function
        compress=_not_implemented,
        decompress=_not_implemented,
        train=_train_zstd_factory(),
    )
    YUNA_COMPRESS_CACHE[COMPRESS_ZSTD] = plugins


_COMPRESS_IMPORT_FUNCTIONS = {
    COMPRESS_LZ4: _import_lz4,
    COMPRESS_ZLIB: _import_zlib,
    COMPRESS_ZSTD: _import_zstd,
}


def get_serialize_plugins(tag: Optional[str]) -> dict:
    if tag is None:
        return None
    plugins = YUNA_SERIALIZE_CACHE.get(tag, None)
    if plugins is None:
        fn_import = _SERIALIZE_IMPORT_FUNCTIONS.get(tag, None)
        if fn_import is None:
            raise ValueError(f"'{tag}': unknown serialization format")
        fn_import()
    plugins = YUNA_SERIALIZE_CACHE.get(tag, None)
    if plugins is None:
        # This error should be impossible...
        # If it happens, check the import function and make sure it saves to the cache with the correct tag.
        raise RuntimeError(f"'{tag}': serialize import succeeded but cannot get plugins")
    return plugins

def get_compress_plugins(tag: str):
    if tag is None:
        return None
    plugins = YUNA_COMPRESS_CACHE.get(tag, None)
    if plugins is None:
        fn_import = _COMPRESS_IMPORT_FUNCTIONS.get(tag, None)
        if fn_import is None:
            raise ValueError(f"'{tag}': unknown compression format")
        fn_import()
    plugins = YUNA_COMPRESS_CACHE.get(tag, None)
    if plugins is None:
        # This error should be impossible...
        # If it happens, check the import function and make sure it saves to the cache with the correct tag.
        raise RuntimeError(f"'{tag}': compress import succeeded but cannot get plugins")
    return plugins


# Python strings are much more convenient as keys than Python byte strings.
# So, while it's legal to use byte strings as keys, Yuna doesn't really
# expect that case, so we will always have a key serialization function.
# This function does nothing, very quickly, to handle that case.  If the
# user is passing a byte string anyway we can return it unchanged.
def _return_bytes_unchanged(x: bytes) -> bytes:
    """
    Return a byte string unchanged.

    Used as a key serialization function in cases where
    no serialization is requested.
    """
    return x


def _get_table_raw_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str]
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged

    def get(self, key: str, default: Optional[bytes]=_YUNA_NOT_PROVIDED) -> Optional[bytes]:
        bytes_key = fn_key_serialize(key)
        result = _lmdb_table_get(env, table, bytes_key, None)
        if result is None:
            if default is _YUNA_NOT_PROVIDED:
                raise KeyError(key)
            else:
                return default
        return result
    return get

def _get_table_deserialize_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_serialize_tag: str
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged

    value_plugins = get_serialize_plugins(value_serialize_tag)
    fn_value_deserialize = value_plugins.deserialize

    def get(self, key: str, default: object=_YUNA_NOT_PROVIDED) -> object:
        bytes_key = fn_key_serialize(key)
        result = _lmdb_table_get(env, table, bytes_key, None)
        if result is None:
            if default is _YUNA_NOT_PROVIDED:
                raise KeyError(key)
            else:
                return default
        return fn_value_deserialize(result)
    return get

def _get_table_decompress_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_compress_tag: str
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged

    value_plugins = get_compress_plugins(value_compress_tag)
    fn_value_decompress = value_plugins.decompress

    def get(self, key: str, default: object=_YUNA_NOT_PROVIDED) -> object:
        bytes_key = fn_key_serialize(key)
        result = _lmdb_table_get(env, table, bytes_key, None)
        if result is None:
            if default is _YUNA_NOT_PROVIDED:
                raise KeyError(key)
            else:
                return default
        return fn_value_decompress(result)
    return get

def _get_table_deserialize_decompress_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_serialize_tag: str,
    value_compress_tag: str
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged

    value_serialize_plugins = get_serialize_plugins(value_serialize_tag)
    fn_value_deserialize = value_serialize_plugins.deserialize

    value_compress_plugins = get_compress_plugins(value_compress_tag)
    fn_value_decompress = value_compress_plugins.decompress

    def get(self, key: str, default: object=_YUNA_NOT_PROVIDED) -> object:
        bytes_key = fn_key_serialize(key)
        result = _lmdb_table_get(env, table, bytes_key, None)
        if result is None:
            if default is _YUNA_NOT_PROVIDED:
                raise KeyError(key)
            else:
                return default
        return fn_value_deserialize(fn_value_decompress(result))
    return get

def get_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_serialize_tag: Optional[str],
    value_compress_tag: Optional[str]
) -> Callable:
    if value_serialize_tag and value_compress_tag:
        return _get_table_deserialize_decompress_factory(
                env, table, key_serialize_tag, value_serialize_tag, value_compress_tag)
    elif not value_serialize_tag and value_compress_tag:
        return _get_table_decompress_factory(
                env, table, key_serialize_tag, value_compress_tag)
    elif value_serialize_tag and not value_compress_tag:
        return _get_table_deserialize_factory(
                env, table, key_serialize_tag, value_serialize_tag)
    else:
        return _get_table_raw_factory(env, table, key_serialize_tag)


def _put_table_raw_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str]
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged

    def put(self, key: str, bytes_value: bytes) -> None:
        bytes_key = fn_key_serialize(key)
        _lmdb_table_put(env, table, bytes_key, bytes_value)
    return put

def _put_table_serialize_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_serialize_tag: str
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged

    value_plugins = get_serialize_plugins(value_serialize_tag)
    fn_value_serialize = value_plugins.serialize

    def put(self, key: str, value: object) -> object:
        bytes_key = fn_key_serialize(key)
        bytes_value = fn_value_serialize(value)
        _lmdb_table_put(env, table, bytes_key, bytes_value)
    return put

def _put_table_compress_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_compress_tag: str
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged

    value_plugins = get_compress_plugins(value_compress_tag)
    fn_value_compress = value_plugins.compress

    def put(self, key: str, value: object) -> object:
        bytes_key = fn_key_serialize(key)
        bytes_value = fn_value_compress(result)
        _lmdb_table_put(env, table, bytes_key, bytes_value)
    return put

def _put_table_serialize_compress_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_serialize_tag: str,
    value_compress_tag: str
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged

    value_serialize_plugins = get_serialize_plugins(value_serialize_tag)
    fn_value_serialize = value_serialize_plugins.serialize

    value_compress_plugins = get_compress_plugins(value_compress_tag)
    fn_value_compress = value_compress_plugins.compress

    def put(self, key: str, value: object) -> object:
        bytes_key = fn_key_serialize(key)
        bytes_value = fn_value_compress(fn_value_serialize(value))
        _lmdb_table_put(env, table, bytes_key, bytes_value)
    return put

def put_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_serialize_tag: Optional[str],
    value_compress_tag: Optional[str]
) -> Callable:
    if value_serialize_tag and value_compress_tag:
        return _put_table_serialize_compress_factory(
                env, table, key_serialize_tag, value_serialize_tag, value_compress_tag)
    elif not value_serialize_tag and value_compress_tag:
        return _put_table_compress_factory(
                env, table, key_serialize_tag, value_compress_tag)
    elif value_serialize_tag and not value_compress_tag:
        return _put_table_serialize_factory(
                env, table, key_serialize_tag, value_serialize_tag)
    else:
        return _put_table_raw_factory(env, table, key_serialize_tag)

def delete_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str]
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged

    def delete(self, key: str) -> None:
        bytes_key = fn_key_serialize(key)
        _lmdb_table_delete(env, table, bytes_key)
    return delete


def keys_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str]
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged
    fn_key_deserialize = key_plugins.deserialize if key_plugins else _return_bytes_unchanged

    def keys(self, start: Optional[str]=None, stop: Optional[str]=None) -> Iterator:
        _empty_string_key_check(start)
        _empty_string_key_check(stop)
        bytes_start = fn_key_serialize(start) if (start is not None) else None
        bytes_stop = fn_key_serialize(stop) if (stop is not None) else None
        with env.begin() as txn:
            with txn.cursor(table) as cursor:
                if bytes_start is not None:
                    cursor.set_range(bytes_start)
                if bytes_stop is None:
                    for bytes_key, _ in cursor:
                        key = fn_key_deserialize(bytes_key)
                        yield key
                else:
                    for bytes_key, _ in cursor:
                        if bytes_key >= bytes_stop:
                            break
                        key = fn_key_deserialize(bytes_key)
                        yield key
    return keys


def _items_table_raw_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str]
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged
    fn_key_deserialize = key_plugins.deserialize if key_plugins else _return_bytes_unchanged

    def items(self, start: Optional[str]=None, stop: Optional[str]=None) -> Iterator:
        _empty_string_key_check(start)
        _empty_string_key_check(stop)
        bytes_start = fn_key_serialize(start) if (start is not None) else None
        bytes_stop = fn_key_serialize(stop) if (stop is not None) else None
        with env.begin() as txn:
            with txn.cursor(table) as cursor:
                if bytes_start is not None:
                    cursor.set_range(bytes_start)
                if bytes_stop is None:
                    if fn_key_deserialize is _return_bytes_unchanged:
                        # The very fastest possible case: return byte keys and byte values, just use yield from!
                        yield from cursor
                    else:
                        for bytes_key, bytes_value in cursor:
                            key = fn_key_deserialize(bytes_key)
                            yield key, bytes_value
                else:
                    for bytes_key, bytes_value in cursor:
                        if bytes_key >= bytes_stop:
                            break
                        key = fn_key_deserialize(bytes_key)
                        yield key, bytes_value
    return items

def _items_table_deserialize_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_serialize_tag: str
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged
    fn_key_deserialize = key_plugins.deserialize if key_plugins else _return_bytes_unchanged

    value_plugins = get_serialize_plugins(value_serialize_tag)
    fn_value_deserialize = value_plugins.deserialize

    def items(self, start: Optional[str]=None, stop: Optional[str]=None) -> Iterator:
        _empty_string_key_check(start)
        _empty_string_key_check(stop)
        bytes_start = fn_key_serialize(start) if (start is not None) else None
        bytes_stop = fn_key_serialize(stop) if (stop is not None) else None
        with env.begin() as txn:
            with txn.cursor(table) as cursor:
                if bytes_start is not None:
                    cursor.set_range(bytes_start)
                if bytes_stop is None:
                    for bytes_key, bytes_value in cursor:
                        key = fn_key_deserialize(bytes_key)
                        value = fn_value_deserialize(bytes_value)
                        yield key, value
                else:
                    for bytes_key, bytes_value in cursor:
                        if bytes_key >= bytes_stop:
                            break
                        key = fn_key_deserialize(bytes_key)
                        value = fn_value_deserialize(bytes_value)
                        yield key, value
    return items

def _items_table_decompress_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_compress_tag: str
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged
    fn_key_deserialize = key_plugins.deserialize if key_plugins else _return_bytes_unchanged

    value_plugins = get_compress_plugins(value_compress_tag)
    fn_value_decompress = value_plugins.decompress

    def items(self, start: Optional[str]=None, stop: Optional[str]=None) -> Iterator:
        _empty_string_key_check(start)
        _empty_string_key_check(stop)
        bytes_start = fn_key_serialize(start) if (start is not None) else None
        bytes_stop = fn_key_serialize(stop) if (stop is not None) else None
        with env.begin() as txn:
            with txn.cursor(table) as cursor:
                if bytes_start is not None:
                    cursor.set_range(bytes_start)
                if bytes_stop is None:
                    for bytes_key, bytes_value in cursor:
                        key = fn_key_deserialize(bytes_key)
                        value = fn_value_decompress(bytes_value)
                        yield key, value
                else:
                    for bytes_key, bytes_value in cursor:
                        if bytes_key >= bytes_stop:
                            break
                        key = fn_key_deserialize(bytes_key)
                        value = fn_value_decompress(bytes_value)
                        yield key, value
    return items

def _items_table_deserialize_decompress_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_serialize_tag: str,
    value_compress_tag: str
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged
    fn_key_deserialize = key_plugins.deserialize if key_plugins else _return_bytes_unchanged

    value_serialize_plugins = get_serialize_plugins(value_serialize_tag)
    fn_value_deserialize = value_serialize_plugins.deserialize

    value_compress_plugins = get_compress_plugins(value_compress_tag)
    fn_value_decompress = value_compress_plugins.decompress

    def items(self, start: Optional[str]=None, stop: Optional[str]=None) -> Iterator:
        _empty_string_key_check(start)
        _empty_string_key_check(stop)
        bytes_start = fn_key_serialize(start) if (start is not None) else None
        bytes_stop = fn_key_serialize(stop) if (stop is not None) else None
        with env.begin() as txn:
            with txn.cursor(table) as cursor:
                if bytes_start is not None:
                    cursor.set_range(bytes_start)
                if bytes_stop is None:
                    for bytes_key, bytes_value in cursor:
                        key = fn_key_deserialize(bytes_key)
                        value = fn_value_deserialize(fn_value_decompress(bytes_value))
                        yield key, value
                else:
                    for bytes_key, bytes_value in cursor:
                        if bytes_key >= bytes_stop:
                            break
                        key = fn_key_deserialize(bytes_key)
                        value = fn_value_deserialize(fn_value_decompress(bytes_value))
                        yield key, value
    return items

def items_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_serialize_tag: Optional[str],
    value_compress_tag: Optional[str]
) -> Callable:
    if value_serialize_tag and value_compress_tag:
        return _items_table_deserialize_decompress_factory(
                env, table, key_serialize_tag, value_serialize_tag, value_compress_tag)
    elif not value_serialize_tag and value_compress_tag:
        return _items_table_decompress_factory(
                env, table, key_serialize_tag, value_compress_tag)
    elif value_serialize_tag and not value_compress_tag:
        return _items_table_deserialize_factory(
                env, table, key_serialize_tag, value_serialize_tag)
    else:
        return _items_table_raw_factory(env, table, key_serialize_tag)


def _values_table_raw_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str]
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged

    def values(self, start: Optional[str]=None, stop: Optional[str]=None) -> Iterator:
        _empty_string_key_check(start)
        _empty_string_key_check(stop)
        bytes_start = fn_key_serialize(start) if (start is not None) else None
        bytes_stop = fn_key_serialize(stop) if (stop is not None) else None
        with env.begin() as txn:
            with txn.cursor(table) as cursor:
                if bytes_start is not None:
                    cursor.set_range(bytes_start)
                if bytes_stop is None:
                    for _, bytes_value in cursor:
                        yield bytes_value
                else:
                    for bytes_key, bytes_value in cursor:
                        if bytes_key >= bytes_stop:
                            break
                        yield bytes_value
    return values

def _values_table_deserialize_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_serialize_tag: str
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged

    value_plugins = get_serialize_plugins(value_serialize_tag)
    fn_value_deserialize = value_plugins.deserialize

    def values(self, start: Optional[str]=None, stop: Optional[str]=None) -> Iterator:
        _empty_string_key_check(start)
        _empty_string_key_check(stop)
        bytes_start = fn_key_serialize(start) if (start is not None) else None
        bytes_stop = fn_key_serialize(stop) if (stop is not None) else None
        with env.begin() as txn:
            with txn.cursor(table) as cursor:
                if bytes_start is not None:
                    cursor.set_range(bytes_start)
                if bytes_stop is None:
                    for _, bytes_value in cursor:
                        value = fn_value_deserialize(bytes_value)
                        yield value
                else:
                    for bytes_key, bytes_value in cursor:
                        if bytes_key >= bytes_stop:
                            break
                        value = fn_value_deserialize(bytes_value)
                        yield value
    return values

def _values_table_decompress_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_compress_tag: str
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged

    value_plugins = get_compress_plugins(value_compress_tag)
    fn_value_decompress = value_plugins.decompress

    def values(self, start: Optional[str]=None, stop: Optional[str]=None) -> Iterator:
        _empty_string_key_check(start)
        _empty_string_key_check(stop)
        bytes_start = fn_key_serialize(start) if (start is not None) else None
        bytes_stop = fn_key_serialize(stop) if (stop is not None) else None
        with env.begin() as txn:
            with txn.cursor(table) as cursor:
                if bytes_start is not None:
                    cursor.set_range(bytes_start)
                if bytes_stop is None:
                    for _, bytes_value in cursor:
                        value = fn_value_decompress(bytes_value)
                        yield value
                else:
                    for bytes_key, bytes_value in cursor:
                        if bytes_key >= bytes_stop:
                            break
                        value = fn_value_decompress(bytes_value)
                        yield value
    return values

def _values_table_deserialize_decompress_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_serialize_tag: str,
    value_compress_tag: str
) -> Callable:
    key_plugins = get_serialize_plugins(key_serialize_tag)
    fn_key_serialize = key_plugins.serialize if key_plugins else _return_bytes_unchanged

    value_serialize_plugins = get_serialize_plugins(value_serialize_tag)
    fn_value_deserialize = value_serialize_plugins.deserialize

    value_compress_plugins = get_compress_plugins(value_compress_tag)
    fn_value_decompress = value_compress_plugins.decompress

    def values(self, start: Optional[str]=None, stop: Optional[str]=None) -> Iterator:
        _empty_string_key_check(start)
        _empty_string_key_check(stop)
        bytes_start = fn_key_serialize(start) if (start is not None) else None
        bytes_stop = fn_key_serialize(stop) if (stop is not None) else None
        with env.begin() as txn:
            with txn.cursor(table) as cursor:
                if bytes_start is not None:
                    cursor.set_range(bytes_start)
                if bytes_stop is None:
                    for _, bytes_value in cursor:
                        value = fn_value_deserialize(fn_value_decompress(bytes_value))
                        yield value
                else:
                    for bytes_key, bytes_value in cursor:
                        if bytes_key >= bytes_stop:
                            break
                        value = fn_value_deserialize(fn_value_decompress(bytes_value))
                        yield value
    return values

def values_factory(
    env: lmdb.Environment,
    table: lmdb._Database,
    key_serialize_tag: Optional[str],
    value_serialize_tag: Optional[str],
    value_compress_tag: Optional[str]
) -> Callable:
    if value_serialize_tag and value_compress_tag:
        return _values_table_deserialize_decompress_factory(
                env, table, key_serialize_tag, value_serialize_tag, value_compress_tag)
    elif not value_serialize_tag and value_compress_tag:
        return _values_table_decompress_factory(
                env, table, key_serialize_tag, value_compress_tag)
    elif value_serialize_tag and not value_compress_tag:
        return _values_table_deserialize_factory(
                env, table, key_serialize_tag, value_serialize_tag)
    else:
        return _values_table_raw_factory(env, table, key_serialize_tag)



# -*- coding: utf-8 -*-

"""
Yuna DB: dict-like semantics for LMDB

Yuna is a key/value store that's built upon LMDB (the Symas Lightning
Memory-mapped Database).  LMDB really is lightning-fast, but as a C
library only allows you to lookup a byte string value, using either
an integer key or a byte string key.

Yuna provides semantics similar to a dict.  You can specify a serialization
format, and optionally a compression format to use, and Yuna will serialize
values and write them to the database as byte strings.  It will
automatically recover the value from the bytes string.

For example, if x is any Python value that can be serialized by the chosen
serialization format in table foo, this would work:

db.tables.foo.put(key, x)

After that .put() you can call .get():

x = db.tables.foo.get(key)
"""

__version__ = "0.2.2"

import os
import types

from typing import Any, Iterator, Optional

import lmdb


from .lmdb_util import YUNA_DEFAULT_MAX_DB_SIZE, YUNA_DEFAULT_MAX_TABLES
from .lmdb_util import YUNA_DB_META_KEY, YUNA_FILE_EXTENSION

from .lmdb_util import _lmdb_open, _yuna_get_meta, _yuna_new_meta, _yuna_put_meta
from .lmdb_util import _lmdb_reserved_delete, _lmdb_reserved_get, _lmdb_reserved_put
from .lmdb_util import _lmdb_table_drop, _lmdb_table_open, _lmdb_table_truncate

from . import plugins
from .plugins import _YUNA_NOT_PROVIDED
from .plugins import SERIALIZE_JSON, SERIALIZE_MSGPACK, SERIALIZE_STR
from .plugins import COMPRESS_LZ4, COMPRESS_ZLIB, COMPRESS_ZSTD
from .plugins import serialize_json, deserialize_json
from .plugins import serialize_str, deserialize_str
from .plugins import _empty_string_key_check


class YunaSharedData:
    """
    Private data for Yuna, in a class by itself so it can be shared
    among the multiple classes implementing Yuna.
    """
    def __init__(self, env: lmdb.Environment, tables_map: dict, metadata: dict):
        self.env = env
        self.tables_map = tables_map
        self.metadata = metadata


class YunaReservedTable:
    """
    This class provides method functions to get/put values from the
    LMDB reserved table.

    This will be .reserved in the open Yuna instance.
    """
    # This is opinionated code.  You cannot specify a serialization or compression format.
    # If you need to do anything that requires a specific serialization or compression
    # format, create a table and use that.  The reserved table should be mostly left alone.
    # Note that LMDB stores things in the reserved table, and Bad Things would happen if
    # you clobbered one of their special values.  In particular, any name used for a table
    # must not be clobbered.
    #
    # If you somehow have a real need to put something other than JSON into the reserved
    # table, serialize it yourself and use .raw_put() to store it.
    #
    # LMDB lets you have any number of tables; use those and leave the reserved table alone.

    def __init__(self, env: lmdb.Environment):
        self.env = env

    def delete(self, key: str) -> None:
        bytes_key = serialize_str(key)
        _lmdb_reserved_delete(self.env, bytes_key)

    def get(self, key: str, default: Any=_YUNA_NOT_PROVIDED) -> Any:
        bytes_key = serialize_str(key)
        bytes_value = _lmdb_reserved_get(self.env, bytes_key, None)
        if bytes_value is None:
            if default is _YUNA_NOT_PROVIDED:
                raise KeyError(key)
            else:
                return default
        value = deserialize_json(bytes_value)
        return value

    def put(self, key: str, value: Any) -> None:
        bytes_key = serialize_str(key)
        bytes_value = serialize_json(value)
        _lmdb_reserved_put(self.env, bytes_key, bytes_value)

    def keys(self, start: Optional[str]=None, stop: Optional[str]=None) -> Iterator:
        _empty_string_key_check(start)
        _empty_string_key_check(stop)
        bytes_start = serialize_str(start) if (start is not None) else None
        bytes_stop = serialize_str(stop) if (stop is not None) else None
        with self.env.begin() as txn:
            with txn.cursor() as cursor:
                if bytes_start is not None:
                    cursor.set_range(bytes_start)
                if bytes_stop is None:
                    for bytes_key, _ in cursor:
                        key = deserialize_str(bytes_key)
                        yield key
                else:
                    for bytes_key, _ in cursor:
                        if bytes_key >= bytes_stop:
                            break
                        key = deserialize_str(bytes_key)
                        yield key

    def raw_delete(self, bytes_key: bytes) -> None:
        _lmdb_reserved_delete(self.env, bytes_key)

    def raw_get(self, bytes_key: bytes, default: Any=_YUNA_NOT_PROVIDED) -> Any:
        bytes_value = _lmdb_reserved_get(self.env, bytes_key, None)
        if bytes_value is None:
            if default is _YUNA_NOT_PROVIDED:
                raise KeyError(key)
            else:
                return default
        return bytes_value

    def raw_put(self, bytes_key: bytes, bytes_value: bytes) -> None:
        _lmdb_reserved_put(self.env, bytes_key, bytes_value)

    def raw_keys(self, bytes_start: Optional[bytes]=None, bytes_stop: Optional[bytes]=None) -> Iterator:
        _empty_string_key_check(bytes_start)
        _empty_string_key_check(bytes_stop)
        with self.env.begin() as txn:
            with txn.cursor() as cursor:
                if bytes_start is not None:
                    cursor.set_range(bytes_start)
                if bytes_stop is None:
                    for bytes_key, _ in cursor:
                        yield bytes_key
                else:
                    for bytes_key, _ in cursor:
                        if bytes_key >= bytes_stop:
                            break
                        yield bytes_key

    def raw_items(self, bytes_start: Optional[bytes]=None, bytes_stop: Optional[bytes]=None) -> Iterator:
        _empty_string_key_check(bytes_start)
        _empty_string_key_check(bytes_stop)
        with self.env.begin() as txn:
            with txn.cursor() as cursor:
                if bytes_start is not None:
                    cursor.set_range(bytes_start)
                if bytes_stop is None:
                    for bytes_key, bytes_value in cursor:
                        yield bytes_key, bytes_value
                else:
                    for bytes_key, bytes_value in cursor:
                        if bytes_key >= bytes_stop:
                            break
                        yield bytes_key, bytes_value

    def raw_values(self, bytes_start: Optional[bytes]=None, bytes_stop: Optional[bytes]=None) -> Iterator:
        _empty_string_key_check(bytes_start)
        _empty_string_key_check(bytes_stop)
        with self.env.begin() as txn:
            with txn.cursor() as cursor:
                if bytes_start is not None:
                    cursor.set_range(bytes_start)
                if bytes_stop is None:
                    for _, bytes_value in cursor:
                        yield bytes_value
                else:
                    for bytes_key, bytes_value in cursor:
                        if bytes_key >= bytes_stop:
                            break
                        yield bytes_value


class YunaTablesMap:
    """
    A trvial class, just used as a container for instances of YunaTable.

    This will be .tables in the open Yuna instance.
    """
    pass


class YunaTableMetadata:
    def __init__(self,
        name: str,
        key_serialize: Optional[str] = None,
        serialize: Optional[str] = None,
        compress: Optional[str] = None,
    ):
        self.name = name
        self.key_serialize = key_serialize
        self.serialize = serialize
        self.compress = compress


class YunaTable:
    """
    This class implements a table for Yuna.

    Provides method functions for delete, get, put, etc.
    """
    def __init__(self,
        shared: YunaSharedData,
        name: str,
        key_serialize: Optional[str],
        serialize: Optional[str],
        compress: Optional[str],
    ):
        if name in shared.tables_map:
            raise ValueError(f"table '{name}' is already open in this database")

        # TODO: check key_serialize to see if we are doing integer keys here
        try:
            temp = plugins.get_serialize_plugins(key_serialize)
        except ValueError:
            raise ValueError("unknown serialization format for key_serialize: {repr(key_serialize)}")
        try:
            temp = plugins.get_serialize_plugins(serialize)
        except ValueError:
            raise ValueError("unknown serialization format for serialize: {repr(serialize)}")
        try:
            temp = plugins.get_compress_plugins(compress)
        except ValueError:
            raise ValueError("unknown compression format for compress: {repr(compress)}")

        meta = YunaTableMetadata(
            name=name,
            key_serialize=key_serialize, serialize=serialize, compress=compress
        )
        self._shared = shared

        # Check to see if the table name is in the metadata.  If it is in there, assume the table exists
        # in the LMDB file, so we wouldn't want to create.  If it's not in there, we need to create it.
        create = name not in self._shared.metadata["tables"]

        # integerkey forced false for now
        self.lmdb_table = _lmdb_table_open(self._shared.env, name, create=create, integerkey=False)

        self.name = name
        self.meta = meta

        key_serialize = meta.key_serialize

        # add method functions based on what's documented in the metadata
        env = self._shared.env
        lmdb_table = self.lmdb_table

        temp = plugins.delete_factory(env, lmdb_table, key_serialize)
        self.delete = types.MethodType(temp, self)
        temp = plugins.delete_factory(env, lmdb_table, None)
        temp.__name__ = "raw_delete"
        self.raw_delete = types.MethodType(temp, self)

        temp = plugins.get_factory(env, lmdb_table, key_serialize, serialize, compress)
        self.get = types.MethodType(temp, self)
        temp = plugins.get_factory(env, lmdb_table, None, None, None)
        temp.__name__ = "raw_get"
        self.raw_get = types.MethodType(temp, self)

        temp = plugins.put_factory(env, lmdb_table, key_serialize, serialize, compress)
        self.put = types.MethodType(temp, self)
        temp = plugins.put_factory(env, lmdb_table, None, None, None)
        temp.__name__ = "raw_put"
        self.raw_put = types.MethodType(temp, self)

        temp = plugins.items_factory(env, lmdb_table, key_serialize, serialize, compress)
        self.items = types.MethodType(temp, self)
        temp = plugins.items_factory(env, lmdb_table, None, None, None)
        temp.__name__ = "raw_items"
        self.raw_items = types.MethodType(temp, self)

        temp = plugins.keys_factory(env, lmdb_table, key_serialize)
        self.keys = types.MethodType(temp, self)
        temp = plugins.keys_factory(env, lmdb_table, None)
        temp.__name__ = "raw_keys"
        self.raw_keys = types.MethodType(temp, self)

        temp = plugins.values_factory(env, lmdb_table, key_serialize, serialize, compress)
        self.values = types.MethodType(temp, self)
        temp = plugins.values_factory(env, lmdb_table, None, None, None)
        temp.__name__ = "raw_values"
        self.raw_values = types.MethodType(temp, self)

        # Table instance fully created so keep track of it
        self._shared.tables_map[name] = self
        if create:
            self._shared.metadata["tables"][name] = vars(self.meta)
            _yuna_put_meta(self._shared.env, self._shared.metadata)
        else:
            assert self._shared.metadata["tables"][name] == vars(self.meta)

    def drop(self):
        _lmdb_table_drop(self._shared.env, self.lmdb_table)
        del self._shared.tables_map[self.name]
        del self._shared.metadata["tables"][self.name]
        self._shared = self.lmdb_table = self.name = self.meta = None

    def truncate(self):
        _lmdb_table_truncate(self._shared.env, self.lmdb_table)


class Yuna:
    """
    Key/value store with dict-like semantics.  A wrapper around LMDB.
    """
    def __init__(self,
        fname: str,
        # YunaDB name and version
        name: Optional[str]=None,
        version: Optional[int]=None,

        # details of an LMDB file follow
        read_only: bool=True,
        create: bool=False,
        safety_mode: str='a',
        single_file: bool=True,
        max_tables: int=YUNA_DEFAULT_MAX_TABLES,
        max_db_file_size: int=YUNA_DEFAULT_MAX_DB_SIZE,
        **kwargs
    ):
        if create:
            read_only = False

        env = _lmdb_open(fname,
                read_only=read_only, create=create, safety_mode=safety_mode, single_file=single_file,
                max_tables=max_tables, max_db_file_size=max_db_file_size,
                extra_args=kwargs)
        if create:
            metadata = _yuna_new_meta(name=name, version=version)
        else:
            metadata = _yuna_get_meta(env, name=name, version=version)

        tables = YunaTablesMap()
        reserved = YunaReservedTable(env=env)

        self.pathname = os.path.abspath(fname)
        self.metadata = metadata
        self.reserved = reserved
        self._shared = YunaSharedData(env=env, tables_map=vars(tables), metadata=metadata)
        self.tables = tables

        # Set up an entry in .tables for each table listed in metadata, with delete/get/put functions ready to use.
        for meta in metadata["tables"].values():
            YunaTable(self._shared, meta["name"], meta["key_serialize"], meta["serialize"], meta["compress"])

    def __enter__(self):
        return self
    def __exit__(self, exc_type, exc_value, traceback):
        self._shared.env.close()
        return False # if there was an exception, do not suppress it

    def sync(self):
        """
        Ensure that all data is flushed to disk.

        Useful when Yuna was opened in "unsafe" mode.
        """
        _lmdb_sync(self._shared.env)
    def close(self):
        """
        Close the Yuna instance.

        Ensures that all data is flushed to disk.
        """
        self._shared.env.close()
        del self.tables
        del self._shared
    @property
    def table_names(self):
        return sorted(vars(self.tables))
    def new_table(self,
        name: str,
        key_serialize: Optional[str]=SERIALIZE_STR,
        serialize: Optional[str]=None,
        compress: Optional[str]=None
    ):
        """
        Open a new Yuna table.

        Creates the table in the LMDB file, updates the Yuna metadata,
        and sets up serialization and optional compression as requested.
        """
        tbl = YunaTable(self._shared, name, key_serialize, serialize, compress)

        # YunaTable takes care of adding the new table to self.tables
        assert name in vars(self.tables)
        # YunaTable also updates the metadata
        assert name in self._shared.metadata["tables"]

        return tbl

class YunaReadOnly(Yuna):
    def __init__(self, *args, **kwargs):
        kwargs["read_only"] = True
        kwargs["create"] = False
        kwargs["safety_mode"] = 'u'
        super().__init__(*args, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import tempfile
import unittest


# Hack sys.path so that tests will run against Yuna from this directory tree,
# even if someone ran "pip install yuna-db" before running these unit tests.
THIS_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.abspath(os.path.join(THIS_DIR, "../src")))
import yuna


# Data that can be reused in multiple tests
_TEST_DATA = {
    "foo": {'f': 0, 'o': 1},
    "bar": {'b': 9, 'a': 8, 'r': 7},
}


class YunaDBTestCase(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        self.fname = os.path.join(self.tmpdir.name, "yuna_test.ydb")
        self.db = yuna.Yuna(self.fname, "test", 1, create=True)

    def tearDown(self):
        self.db.close()
        self.tmpdir.cleanup()

    def test_reserved_get_put(self):
        self.db.reserved.put("foo", _TEST_DATA["foo"])
        self.db.reserved.put("bar", _TEST_DATA["bar"])
        tbl_reserved = self.db.reserved

        x = tbl_reserved.get("foo", None)
        self.assertEqual(x, _TEST_DATA["foo"])
        x = tbl_reserved.get("bar", None)
        self.assertEqual(x, _TEST_DATA["bar"])
        x = tbl_reserved.get("baz", None)
        self.assertIsNone(x)

    def test_reserved_delete(self):
        self.db.reserved.put("foo", _TEST_DATA["foo"])
        self.db.reserved.put("bar", _TEST_DATA["bar"])
        tbl_reserved = self.db.reserved

        x = tbl_reserved.get("foo", None)
        self.assertEqual(x, _TEST_DATA["foo"])
        tbl_reserved.delete("foo")
        x = tbl_reserved.get("foo", None)
        self.assertIsNone(x)

        x = tbl_reserved.get("bar", None)
        self.assertEqual(x, _TEST_DATA["bar"])
        tbl_reserved.raw_delete(b"bar")
        x = tbl_reserved.get("bar", None)
        self.assertIsNone(x)

    def test_reserved_keys(self):
        # NOTE: if we don't have any other tables open, reserved will start out empty
        self.db.reserved.put("foo", _TEST_DATA["foo"])
        self.db.reserved.put("bar", _TEST_DATA["bar"])
        tbl_reserved = self.db.reserved

        # keys are listed in sorted order
        x = list(tbl_reserved.keys())
        expected = ["bar", "foo"]
        self.assertEqual(x, expected)

        x = list(tbl_reserved.raw_keys())
        expected = [b"bar", b"foo"]
        self.assertEqual(x, expected)

    def test_reserved_raw_items(self):
        # NOTE: if we don't have any other tables open, reserved will start out empty
        self.db.reserved.put("foo", _TEST_DATA["foo"])
        self.db.reserved.put("bar", _TEST_DATA["bar"])
        tbl_reserved = self.db.reserved

        x = list(tbl_reserved.raw_items())
        expected = [
            (b"bar", yuna.serialize_json(_TEST_DATA["bar"])),
            (b"foo", yuna.serialize_json(_TEST_DATA["foo"]))
        ]
        self.assertEqual(x, expected)

    def test_reserved_raw_values(self):
        # NOTE: if we don't have any other tables open, reserved will start out empty
        self.db.reserved.put("foo", _TEST_DATA["foo"])
        self.db.reserved.put("bar", _TEST_DATA["bar"])
        tbl_reserved = self.db.reserved

        x = list(tbl_reserved.raw_values())
        expected = [
            yuna.serialize_json(_TEST_DATA["bar"]),
            yuna.serialize_json(_TEST_DATA["foo"])
        ]
        self.assertEqual(x, expected)


    def setup_a26(self):
        """
        Create the "a26" table and put data in it that the other tests
        will expect.
        """
        tbl_a26 = self.db.new_table("a26", serialize=yuna.SERIALIZE_STR, compress=yuna.COMPRESS_LZ4)

        tbl_a26.put('a', '1')
        tbl_a26.put('b', '2')
        # skip 'c' on purpose
        tbl_a26.put('d', '4')
        tbl_a26.put('e', '5')

        return tbl_a26

    def test_a26_drop(self):
        tbl_a26 = self.setup_a26()

        self.assertIn("a26", self.db.table_names)
        tbl_a26.drop()
        self.assertNotIn("a26", self.db.table_names)

    def test_a26_get_put(self):
        tbl_a26 = self.setup_a26()

        x = tbl_a26.get('a', None)
        self.assertEqual(x, '1')
        x = tbl_a26.get('b', None)
        self.assertEqual(x, '2')
        x = tbl_a26.get('c', None)
        self.assertIsNone(x)

        import lz4
        import lz4.block
        def a26_pack(x):
            return lz4.block.compress(yuna.serialize_str(x))

        tbl_a26.raw_put(b'c', a26_pack('3'))
        x = tbl_a26.get('c', None)
        self.assertEqual(x, '3')

    def test_a26_delete(self):
        tbl_a26 = self.setup_a26()

        x = tbl_a26.get('a', None)
        self.assertEqual(x, '1')

        tbl_a26.delete('a')
        x = tbl_a26.get('a', None)
        self.assertIsNone(x)

        x = tbl_a26.get('b', None)
        self.assertEqual(x, '2')

        tbl_a26.raw_delete(b'b')
        x = tbl_a26.get('b', None)
        self.assertIsNone(x)

    def test_a26_keys(self):
        tbl_a26 = self.setup_a26()

        x = list(tbl_a26.keys())
        expected = ['a', 'b', 'd', 'e']
        self.assertEqual(x, expected)

        x = list(tbl_a26.raw_keys())
        expected = [b'a', b'b', b'd', b'e']
        self.assertEqual(x, expected)

    def test_a26_items(self):
        import lz4
        import lz4.block
        def a26_unpack(x):
            return yuna.deserialize_str(lz4.block.decompress(x))

        tbl_a26 = self.setup_a26()

        x = list(tbl_a26.items())
        expected = [('a', '1'), ('b', '2'), ('d', '4'), ('e', '5')]
        self.assertEqual(x, expected)

        lst_raw = list(tbl_a26.raw_items())
        lst = [(yuna.deserialize_str(key), a26_unpack(value)) for key, value in lst_raw]
        self.assertEqual(lst, expected)

    def test_a26_values(self):
        import lz4
        import lz4.block
        def a26_unpack(x):
            return yuna.deserialize_str(lz4.block.decompress(x))

        tbl_a26 = self.setup_a26()

        x = list(tbl_a26.values())
        expected = ['1', '2', '4', '5']
        self.assertEqual(x, expected)

        lst_raw = list(tbl_a26.raw_values())
        lst = [a26_unpack(item) for item in lst_raw]
        self.assertEqual(lst, expected)


    def setup_j(self):
        """
        Create the "j" table and put data in it that the other tests
        will expect.
        """
        tbl_j = self.db.new_table("j", serialize=yuna.SERIALIZE_JSON)

        tbl_j.put("foo", _TEST_DATA["foo"])
        tbl_j.put("bar", _TEST_DATA["bar"])

        return tbl_j

    def test_j_drop(self):
        tbl_j = self.setup_j()

        self.assertIn("j", self.db.table_names)
        tbl_j.drop()
        self.assertNotIn("j", self.db.table_names)

    def test_j_get_put(self):
        tbl_j = self.setup_j()

        x = tbl_j.get("foo", None)
        self.assertEqual(x, _TEST_DATA["foo"])
        x = tbl_j.get("bar", None)
        self.assertEqual(x, _TEST_DATA["bar"])
        x = tbl_j.get("baz", None)
        self.assertIsNone(x)

        x = tbl_j.raw_get(b"foo", None)
        self.assertEqual(x, yuna.serialize_json(_TEST_DATA["foo"]))
        x = tbl_j.raw_get(b"bar", None)
        self.assertEqual(x, yuna.serialize_json(_TEST_DATA["bar"]))
        x = tbl_j.raw_get(b"baz", None)
        self.assertIsNone(x)

        BAZ_DATA = {'x': -1, 'y': -2}
        tbl_j.raw_put(b"baz", yuna.serialize_json(BAZ_DATA))
        x = tbl_j.get("baz", None)
        self.assertEqual(x, BAZ_DATA)

    def test_j_delete(self):
        tbl_j = self.setup_j()

        x = tbl_j.get("foo", None)
        self.assertEqual(x, _TEST_DATA["foo"])
        x = tbl_j.delete("foo")
        x = tbl_j.get("foo", None)
        self.assertIsNone(x)

        x = tbl_j.get("bar", None)
        self.assertEqual(x, _TEST_DATA["bar"])
        x = tbl_j.raw_delete(b"bar")
        x = tbl_j.get("bar", None)
        self.assertIsNone(x)

    def test_j_keys(self):
        tbl_j = self.setup_j()

        # keys are listed in sorted order
        x = list(tbl_j.keys())
        expected = ["bar", "foo"]
        self.assertEqual(x, expected)

        x = list(tbl_j.raw_keys())
        expected = [b"bar", b"foo"]
        self.assertEqual(x, expected)

    def test_j_items(self):
        tbl_j = self.setup_j()

        x = list(tbl_j.items())
        expected = [("bar", _TEST_DATA["bar"]), ("foo", _TEST_DATA["foo"])]
        self.assertEqual(x, expected)

        x = list(tbl_j.raw_items())
        expected = [
            (b"bar", yuna.serialize_json(_TEST_DATA["bar"])),
            (b"foo", yuna.serialize_json(_TEST_DATA["foo"]))
        ]
        self.assertEqual(x, expected)

    def test_j_values(self):
        tbl_j = self.setup_j()

        x = list(tbl_j.values())
        expected = [_TEST_DATA["bar"], _TEST_DATA["foo"]]
        self.assertEqual(x, expected)

        x = list(tbl_j.raw_values())
        expected = [
            (yuna.serialize_json(_TEST_DATA["bar"])),
            (yuna.serialize_json(_TEST_DATA["foo"]))
        ]
        self.assertEqual(x, expected)


    def test_yuna_read_only(self):
        pathname = self.db.pathname
        self.setup_a26()
        self.setup_j()
        self.db.close()

        self.db = yuna.YunaReadOnly(pathname, "test", 1)

        x = self.db.tables.a26.get('a', None)
        self.assertEqual(x, '1')

        x = self.db.tables.j.get("foo", None)
        self.assertEqual(x, _TEST_DATA["foo"])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
test_yuna.py -- Test Yuna DB

We need a set of unit tests, but for now we have this.

Exercise several Yuna features to make sure that at least
the basics are working.

"""

import os
import sys

# Hack sys.path so that this file will run against Yuna from this directory tree,
# even if someone ran "pip install yuna-db" before running this.
THIS_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.abspath(os.path.join(THIS_DIR, "src")))

import yuna
from yuna import Yuna
from yuna.lmdb_util import _delete_file_or_dir

TEST_FILE = "/tmp/junk.ydb"

_delete_file_or_dir(TEST_FILE)
with Yuna(TEST_FILE, "test", 1, create=True) as db:
    db.new_table("a26", serialize=yuna.SERIALIZE_STR, compress=yuna.COMPRESS_LZ4)
    tbl_a26 = db.tables.a26
    tbl_a26.put("a", "1")
    tbl_a26.put("b", "2")
    #tbl_a26.put("c", "3")
    tbl_a26.put("d", "4")
    tbl_a26.put("e", "5")

    lst = list(tbl_a26.keys())
    print(f"a26 keys: {lst}")
    lst = list(tbl_a26.keys(start='c'))
    print(f"a26 keys from 'c': {lst}")
    lst = list(tbl_a26.keys(start='c', stop='e'))
    print(f"a26 keys from 'c' to < 'e': {lst}")

    lst = list(tbl_a26.items())
    print(f"a26 items: {lst}")
    lst = list(tbl_a26.items(start='c'))
    print(f"a26 items from 'c': {lst}")
    lst = list(tbl_a26.items(start='c', stop='e'))
    print(f"a26 items from 'c' to < 'e': {lst}")

    lst = list(tbl_a26.values())
    print(f"a26 values: {lst}")
    lst = list(tbl_a26.values(start='c'))
    print(f"a26 values from 'c': {lst}")
    lst = list(tbl_a26.values(start='c', stop='e'))
    print(f"a26 values from 'c' to < 'e': {lst}")

    assert tbl_a26.get("a", None) == "1"
    assert tbl_a26.get("b", None) == "2"
    assert tbl_a26.get("c", None) is None
    tbl_j = db.new_table("jjj", serialize=yuna.SERIALIZE_JSON)
    D_FOO = {"f": 0, "o": 1}
    D_BAR = {"b": 9, "a": 8, "r": 7}
    tbl_j.put("foo", D_FOO)
    tbl_j.put("bar", D_BAR)
    assert tbl_j.get("foo", None) == D_FOO
    assert tbl_j.get("bar", None) == D_BAR
    assert tbl_j.get("baz", None) is None

    db.reserved.put("foo", D_FOO)
    assert db.reserved.get("foo", None) == D_FOO
    temp = db.reserved.raw_get(b"foo")
    db.reserved.raw_put(b"bar", temp)
    temp = db.reserved.get("bar")
    assert temp == D_FOO
    db.reserved.raw_delete(b"bar")
    assert db.reserved.get("bar", None) is None

    print("reserved raw_keys:")
    for bytes_key in db.reserved.raw_keys():
        print(bytes_key)
    print()
    print("reserved raw_items:")
    for bytes_key, bytes_value in db.reserved.raw_items():
        print(bytes_key, "->", bytes_value)
    print()
    print("reserved raw_values:")
    for bytes_value in db.reserved.raw_values():
        print(bytes_value)
    print()

    db.reserved.delete("foo")
    assert db.reserved.get("foo", None) is None

with Yuna(TEST_FILE, "test", 1, read_only=False) as db:
    assert db.tables.a26.get("a", None) == "1"
    assert db.tables.a26.get("b", None) == "2"
    assert db.tables.a26.get("c", None) is None
    assert db.tables.jjj.get("foo", None) == D_FOO
    assert db.tables.jjj.get("bar", None) == D_BAR
    assert db.tables.jjj.get("baz", None) is None
    db.tables.jjj.truncate()
    assert db.tables.jjj.get("foo", None) is None
    assert db.tables.jjj.get("bar", None) is None
    assert db.tables.jjj.get("baz", None) is None
    db.tables.jjj.drop()
    db.tables.a26.drop()



# -*- coding: utf-8 -*-

"""
lmdb_util.py -- LMDB low-level convenience functions

Functions to handle common LMDB operations.

"""


import json
import os
import shutil

from typing import Optional


import lmdb


# Create a unique object used to detect if optional arg not provided.
# Can't use None because user might want to provide None.
_LMDB_UTIL_NOT_PROVIDED = object()


# NOTES about terminology
#
# The LMDB documentation uses some unusual terms.  The Yuna library uses some
# different terms.
#
# An LMDB file may contain multiple key/value stores.  LMDB refers to these
# as "databases" or "named databases".  Yuna will simply call these "tables".
#
# Every LMDB file has one, default key/value store that the user can store
# data in, but that LMDB also stores some data in.  LMDB calls this
# the "unnamed database".  Yuna calls this the "reserved table".
#
# LMDB calls the open LMDB file an "environment" and uses "env" as
# the variable name for what is returned.  Yuna's low-level LMDB code
# uses this terminology but Yuna in general just refers to "the database".

# LMDB files have to be declared with limits: maximum number of tables,
# maximum total file size.  In practice it works quite well to simply give
# very large limit numbers; the database will not actually take up the
# maximum declared size.  On Linux, the file may appear to be that size,
# but in that case the "yuna_repack" utility can be used to make a copy
# of the database file that is minimum size.
#
# TODO: implement the "yuna_repack" utility


class YunaInvalidDB(ValueError):
    pass

YUNA_FILE_EXTENSION = ".ydb"

YUNA_DB_META_KEY = "__YUNA_DB_META__"

YUNA_DEFAULT_MAX_TABLES = 100
YUNA_DEFAULT_MAX_DB_SIZE = 2**40  # one tebibyte (binary terabyte): 1024**4

_VALID_SAFETY_MODES = ('a', 'u')


def _delete_file_or_dir(fname: str):
    # If it doesn't exist, we don't complain, similar to /bin/rm -f on Linux
    if not os.path.exists(fname):
        return

    # Something exists; delete it whether it is a file or a directory.
    try:
        os.remove(fname)
    except IsADirectoryError:
        shutil.rmtree(fname)

def _lmdb_reserved_delete(
    env: lmdb.Environment,
    bytes_key: bytes
) -> None:
    with env.begin(write=True) as txn:
        txn.delete(bytes_key)

def _lmdb_reserved_get(
    env: lmdb.Environment,
    bytes_key: bytes,
    default: Optional[bytes]=_LMDB_UTIL_NOT_PROVIDED
) -> None:
    with env.begin() as txn:
        result = txn.get(bytes_key, default=None)

    if result is None:
        if default is _LMDB_UTIL_NOT_PROVIDED:
            raise KeyError(key)
        else:
            return default
    return result

def _lmdb_reserved_put(
    env: lmdb.Environment,
    bytes_key: bytes,
    bytes_value: bytes
) -> None:
    with env.begin(write=True) as txn:
        txn.put(bytes_key, bytes_value)


def _lmdb_table_open(
    env: lmdb.Environment,
    name: str,
    create: bool=False,
    integerkey: bool=False
) -> None:
    bytes_name = bytes(name, "utf-8")
    return env.open_db(bytes_name, create=create, integerkey=integerkey)


def _lmdb_table_delete(
    env: lmdb.Environment,
    table: lmdb._Database,
    bytes_key: bytes
) -> None:
    with env.begin(db=table, write=True) as txn:
        txn.delete(bytes_key)

def _lmdb_table_drop(
    env: lmdb.Environment,
    table: lmdb._Database
) -> None:
    # With delete=True drops the entire table and all data contained inside it.
    with env.begin(write=True) as txn:
        txn.drop(table, delete=True)

def _lmdb_table_get(
    env: lmdb.Environment,
    table: lmdb._Database,
    key: bytes,
    default: Optional[bytes]=_LMDB_UTIL_NOT_PROVIDED,
) -> Optional[bytes]:
    with env.begin(db=table) as txn:
        result = txn.get(key, default=None)

    if result is None:
        if default is _LMDB_UTIL_NOT_PROVIDED:
            raise KeyError(key)
        else:
            return default
    return result

def _lmdb_table_put(
    env: lmdb.Environment,
    table: lmdb._Database,
    key: bytes,
    value: bytes,
) -> None:
    with env.begin(db=table, write=True) as txn:
        txn.put(key, value)

def _lmdb_table_truncate(
    env: lmdb.Environment,
    table: lmdb._Database
) -> None:
    with env.begin(write=True) as txn:
        # Drops every key/value pair, but with delete=False does not drop the table itself.
        txn.drop(table, delete=False)

def _lmdb_sync(
    env: lmdb.Environment
) -> None:
    env.sync(force=True)



def _yuna_new_meta(
    name: Optional[str]=None,
    version: Optional[str]=None,
    tables_map: Optional[dict]=None,
) -> dict:
    if tables_map is None:
        tables_map = {}

    metadata = {}
    if name is not None:
        metadata["name"] = name
    if version is not None:
        metadata["version"] = version
    metadata["tables"] = {}

    metadata["yuna_version"] = 1

    return metadata

def _yuna_get_meta(
    env: lmdb.Environment,
    name: Optional[str],
    version: Optional[int],
) -> dict:
    """
    Get the metadata stored in a Yuna DB file.

    Given an open LMDB file, first check to see if it's a valid
    Yuna DB file.  If it is, return the metadata as a dict.

    If @name is provided, check to see if the Yuna DB file has that name.
    If @version is provided, check to see if the Yuna DB file has that version.

    Raises YunaInvalidDB on any error.
    """
    # Try to retrieve the metadata, always stored in the reserved table.
    key = bytes(YUNA_DB_META_KEY, "utf-8")
    value = _lmdb_reserved_get(env, key, default=None)
    if value is None:
        fname = env.path()
        raise YunaInvalidDB(f"LMDB file is not a Yuna DB file: {repr(fname)}")

    # We got something... does it decode as valid JSON?
    try:
        meta = json.loads(value)
        if not isinstance(meta, dict):
            # whatever was stored there wasn't a JSON-encoded metadata dictionary
            raise ValueError
    except (ValueError, TypeError, json.decoder.JSONDecodeError):
        fname = env.path()
        raise YunaInvalidDB(f"Yuna DB has corrupted metadata: {repr(fname)}")

    # If user provided name and/or version, make appropriate checks.
    if name is not None:
        temp = meta.get("name", None)
        if temp != name:
            raise YunaInvalidDB(f"LMDB file name mismatch: expected {repr(name)}, got {repr(temp)}")
    if version is not None:
        temp = meta.get("version", None)
        if temp != version:
            raise YunaInvalidDB(f"LMDB file version mismatch: expected {version}, got {temp}")
    return meta


def _yuna_put_meta(
    env: lmdb.Environment,
    meta: dict,
) -> None:
    """
    Given @meta, a dictionary containing metadata, write it to the LMDB file.
    """
    # Try to retrieve the metadata, always stored in the reserved table.
    key = bytes(YUNA_DB_META_KEY, "ascii")
    s = json.dumps(meta)
    value = bytes(s, "utf-8")
    _lmdb_reserved_put(env, key, value)


def _lmdb_open(
    fname: str,
    read_only: bool=True,
    create: bool=False,
    safety_mode: str='a',
    single_file: bool=True,
    max_tables: int=YUNA_DEFAULT_MAX_TABLES,
    max_db_file_size: int=YUNA_DEFAULT_MAX_DB_SIZE,
    extra_args: Optional[dict]=None,
):
    """
    @safety_mode: legal values are 'a' (ACID safety)  'u' (unsafe; fastest)
    """
    if safety_mode not in _VALID_SAFETY_MODES:
        mesg = f"safety_mode must be one of {_VALID_SAFETY_MODES} "\
            "but instead was {repr(safety_mode)}"
        raise ValueError(mesg)

    if not create:
        if not os.path.exists(fname) and not fname.endswith(YUNA_FILE_EXTENSION):
            temp = fname + YUNA_FILE_EXTENSION
            if os.path.exists(fname):
                fname = temp

        if not os.path.exists(fname):
            raise FileNotFoundError(fname)

    # Create implies we want to be able to write.  Don't even check it, just make sure read_only is False.
    if create:
        read_only = False
        _delete_file_or_dir(fname)

    try:
        kwargs = {
            "create": create,
            "map_size": max_db_file_size,
            "max_dbs": max_tables,
            "readonly": read_only,
            "subdir": not single_file,
        }

        if safety_mode == 'u':
            # Change all the settings to their fastest and least safe value.
            # This is ideal for creating a DB file that will later be used
            # read-only, and is a terrible idea for a DB that will be "live"
            # and will have data written to it when a service is running.
            kwargs["metasync"] = False
            kwargs["sync"] = False
            kwargs["writemap"] = True
            kwargs["map_async"] = True

        if extra_args:
            kwargs.update(extra_args)
        env = lmdb.open(fname, **kwargs)
        return env
    except Exception:
        # currently Yuna is just letting LMDB exceptions be raised.
        raise
